#include <allegro5/allegro_physfs.h>
#include <physfs.h>
#include "t3f/t3f.h"

#include "../archive_handler.h"

typedef struct
{

	const char * filename;
	ALLEGRO_PATH * temp_path;
  int file_count;
  int extract_pos;
  int extract_index;

} ARCHIVE_HANDLER_DATA;

bool handler_init(void)
{
  ALLEGRO_PATH * pp = NULL;

  pp = al_get_standard_path(ALLEGRO_EXENAME_PATH);
  if(!pp)
  {
    goto fail;
  }
  if(!PHYSFS_init(al_path_cstr(pp, '/')))
  {
    goto fail;
  }
  al_destroy_path(pp);
  return true;

  fail:
  {
    if(pp)
    {
      al_destroy_path(pp);
    }
    return false;
  }
}

void handler_exit(void)
{
  PHYSFS_deinit();
}

static void * open_archive(const char * fn, ALLEGRO_PATH * temp_path)
{
	ARCHIVE_HANDLER_DATA * data;

  if(!PHYSFS_addToSearchPath(fn, 1))
  {
    goto fail;
  }
	data = malloc(sizeof(ARCHIVE_HANDLER_DATA));
  if(!data)
  {
    goto fail;
  }
  data->filename = fn;
  data->temp_path = temp_path;
  data->file_count = 0;
	return data;

  fail:
  {
    if(data)
    {
      free(data);
    }
    PHYSFS_removeFromSearchPath(fn);
    return NULL;
  }
}

static void close_archive(void * data)
{
	ARCHIVE_HANDLER_DATA * archive_data = (ARCHIVE_HANDLER_DATA *)data;

  PHYSFS_removeFromSearchPath(archive_data->filename);
	free(data);
}

static bool count_file(const char * fn, bool isfolder, void * data)
{
	ARCHIVE_HANDLER_DATA * archive_data = (ARCHIVE_HANDLER_DATA *)data;

  if(!isfolder)
  {
    data->file_count++;
  }
}

static bool extract_file(const char * fn, bool isfolder, void * data)
{
	ARCHIVE_HANDLER_DATA * archive_data = (ARCHIVE_HANDLER_DATA *)data;

  if(!isfolder)
  {
    /* extract when we reach the index */
    if(data->extract_pos == data->extract_index)
    {

    }
    data->extract_pos++;
  }
}

static int count_files(void * data)
{
	ARCHIVE_HANDLER_DATA * archive_data = (ARCHIVE_HANDLER_DATA *)data;
	struct archive *a;
	struct archive_entry *entry;
	int r, r2;
	int total = 0;
  ALLEGRO_FS_INTERFACE * old_fs_interface = NULL;
  ALLEGRO_FILE_INTERFACE * old_file_interface = NULL;

  old_fs_interface = al_get_fs_interface();
  old_file_interface = al_get_file_interface();
  al_set_physfs_file_interface();
  t3f_scan_files("/", count_file, false, data);
	a = archive_read_new();
	archive_read_support_filter_all(a);
	archive_read_support_format_all(a);
	r = archive_read_open_filename(a, archive_data->filename, 10240); // Note 1
	if(r == ARCHIVE_OK)
	{
		while(1)
		{
			r2 = archive_read_next_header(a, &entry);
			if(r2 == ARCHIVE_OK)
			{
				total++;
				archive_entry_pathname(entry);
//				archive_read_data_skip(a);  // Note 2
			}
			else
			{
				break;
			}
		}
		archive_read_free(a);  // Note 3
	}

	return total;
}

static const char * get_file(void * data, int index, char * buffer)
{
	ARCHIVE_HANDLER_DATA * archive_data = (ARCHIVE_HANDLER_DATA *)data;
	struct archive *a;
	struct archive_entry *entry;
	int r, r2;
	int total = 0;

	a = archive_read_new();
	archive_read_support_filter_all(a);
	archive_read_support_format_all(a);
	r = archive_read_open_filename(a, archive_data->filename, 10240); // Note 1
	if(r == ARCHIVE_OK)
	{
		while(1)
		{
			r2 = archive_read_next_header(a, &entry);
			if(r2 == ARCHIVE_OK)
			{
				if(total == index)
				{
					strcpy(buffer, archive_entry_pathname(entry));
					break;
				}
				total++;
//				archive_read_data_skip(a);  // Note 2
			}
			else
			{
				break;
			}
		}
		archive_read_free(a);  // Note 3
	}

	return buffer;
}

static int copy_data(struct archive *ar, struct archive *aw)
{
	int r;
	const void *buff;
	size_t size;
	int64_t offset;

	for(;;)
	{
		r = archive_read_data_block(ar, &buff, &size, &offset);
		if(r == ARCHIVE_EOF)
		{
			return (ARCHIVE_OK);
		}
		if(r < ARCHIVE_OK)
		{
			return (r);
		}
		r = archive_write_data_block(aw, buff, size, offset);
		if(r < ARCHIVE_OK)
		{
			fprintf(stderr, "%s\n", archive_error_string(aw));
			return (r);
		}
	}
}

static const char * extract_current_file(struct archive * a, struct archive_entry * entry, ALLEGRO_PATH * temp_path)
{
	struct archive *ext;
	int flags;
	int r;
	ALLEGRO_PATH * out_path = al_clone_path(temp_path);
	if(!out_path)
	{
		return NULL;
	}

	flags = ARCHIVE_EXTRACT_TIME;
	flags |= ARCHIVE_EXTRACT_PERM;
	flags |= ARCHIVE_EXTRACT_ACL;
	flags |= ARCHIVE_EXTRACT_FFLAGS;

	ext = archive_write_disk_new();
	archive_write_disk_set_options(ext, flags);
	archive_write_disk_set_standard_lookup(ext);
	al_set_path_filename(out_path, archive_entry_pathname(entry));
	archive_entry_set_pathname(entry, al_path_cstr(out_path, '/'));
	r = archive_write_header(ext, entry);
    if(r == ARCHIVE_OK)
	{
    	if(archive_entry_size(entry) > 0)
		{
			r = copy_data(a, ext);
		}
		r = archive_write_finish_entry(ext);
	}
	al_destroy_path(out_path);
	return NULL;
}

static const char * extract_file(void * data, int index, char * buffer)
{
	ARCHIVE_HANDLER_DATA * archive_data = (ARCHIVE_HANDLER_DATA *)data;
	struct archive *a;
	struct archive_entry *entry;
	int r, r2;
	int total = 0;
	strcpy(buffer, "");

	a = archive_read_new();
	archive_read_support_filter_all(a);
	archive_read_support_format_all(a);
	r = archive_read_open_filename(a, archive_data->filename, 10240); // Note 1
	if(r == ARCHIVE_OK)
	{
		while(1)
		{
			r2 = archive_read_next_header(a, &entry);
			if(r2 == ARCHIVE_OK)
			{
				if(total == index)
				{
					extract_current_file(a, entry, archive_data->temp_path);
					strcpy(buffer, archive_entry_pathname(entry));
					break;
				}
				total++;
//				archive_read_data_skip(a);  // Note 2
			}
			else
			{
				break;
			}
		}
		archive_read_free(a);  // Note 3
	}

	return buffer;
}

static OMO_ARCHIVE_HANDLER archive_handler;

OMO_ARCHIVE_HANDLER * omo_get_physfs_archive_handler(void)
{
	memset(&archive_handler, 0, sizeof(OMO_ARCHIVE_HANDLER));
	strcpy(archive_handler.id, "PhysFS");
  archive_handler.init = handler_init;
  archive_handler.exit = handler_exit;
	archive_handler.open_archive = open_archive;
	archive_handler.close_archive = close_archive;
	archive_handler.count_files = count_files;
	archive_handler.get_file = get_file;
	archive_handler.extract_file = extract_file;
	omo_archive_handler_add_type(&archive_handler, ".zip");
	omo_archive_handler_add_type(&archive_handler, ".zsn");
//	omo_archive_handler_add_type(&archive_handler, ".vgz");
//	omo_archive_handler_add_type(&archive_handler, ".rar");
//	omo_archive_handler_add_type(&archive_handler, ".rsn");
	return &archive_handler;
}
